import EventEmitter from 'events';
import { CoreClient } from './client.js';
import { core_config } from '../config.js';
import { check_process, spawn_process } from './cmd.js';
import { ensure_file, ensure_path } from './util.js';
import * as CONST from './const.js';
const { FALLBACK_FEE, FAUCET_MIN_BAL, INIT_BLOCK_CT, SAT_MULTI, RANDOM_PORT } = CONST;
export class CoreDaemon extends EventEmitter {
    constructor(config) {
        super();
        const opt = core_config(config);
        const { isolated } = opt;
        if (isolated) {
            const port = RANDOM_PORT();
            opt.peer_port = port;
            opt.rpc_port = port + 1;
        }
        this._client = new CoreClient(this, opt);
        this._closing = false;
        this._faucet = null;
        this._ready = false;
        this.tasks = [];
        this.params = [
            `-chain=${opt.network}`,
            `-fallbackfee=${FALLBACK_FEE / SAT_MULTI}`,
            `-port=${opt.peer_port}`,
            `-rpcport=${opt.rpc_port}`,
            '-txindex',
            ...opt.params,
            ...opt.core_params
        ];
        if (opt.confpath !== undefined) {
            this.params.push(`-conf=${opt.confpath}`);
        }
        if (opt.datapath !== undefined) {
            this.params.push(`-datadir=${opt.datapath}`);
        }
        process.once('uncaughtException', async (err) => {
            console.log('[core] Daemon caught an error, exiting...');
            await this.shutdown();
            if (opt.debug)
                console.dir(err, { depth: null });
            throw err;
        });
        process.once('unhandledRejection', async (reason) => {
            console.log('[core] Daemon caught a promise rejection, exiting...');
            await this.shutdown();
            throw new Error(String(reason));
        });
        this._opt = opt;
    }
    get client() {
        return this._client;
    }
    get faucet() {
        if (this._faucet === null) {
            throw new Error('Faucet wallet is not loaded!');
        }
        return this._faucet;
    }
    get opt() {
        return this._opt;
    }
    get ready() {
        return this._ready;
    }
    on(event, cb) {
        return super.on(event, cb);
    }
    once(event, cb) {
        return super.once(event, cb);
    }
    emit(event, arg) {
        return super.emit(event, arg);
    }
    async _start(params = []) {
        const { confpath, corepath, datapath, debug, timeout } = this.opt;
        if (confpath !== undefined) {
            await ensure_file(confpath);
        }
        if (datapath !== undefined) {
            await ensure_path(datapath);
        }
        const exec = corepath ?? 'bitcoind';
        const msg = 'loadblk thread exit';
        params = [...this.params, ...params];
        if (debug) {
            console.log('[core] exec :', exec);
            console.log('[core] data :', datapath);
            console.log('[core] args :', params.join(' '));
        }
        this._proc = await spawn_process(exec, params, msg, timeout);
    }
    async _init() {
        const debug = this.opt.debug;
        const min_bal = FAUCET_MIN_BAL / SAT_MULTI;
        this._faucet = await this.client.load_wallet('faucet');
        if (this.opt.network === 'regtest') {
            const addr = await this.faucet.get_address('faucet');
            let bal = await this.faucet.balance;
            while (bal <= min_bal) {
                if (debug)
                    console.log('[core] Faucet generating blocks. Balance:', bal);
                await this.client.mine_blocks(INIT_BLOCK_CT, addr);
                bal = await this.faucet.balance;
            }
        }
        await Promise.all(this.tasks.map(t => t(this.client)));
    }
    async startup(params = []) {
        const { isolated, verbose } = this.opt;
        if (isolated) {
            await this._start(params);
        }
        else {
            if (await check_process('bitcoin-qt')) {
                if (verbose)
                    console.log('[core]: Using existing bitcoin QT instance...');
            }
            else if (await check_process('bitcoind')) {
                if (verbose)
                    console.log('[core] Using existing bitcoin daemon...');
            }
            else {
                if (verbose)
                    console.log('[core] Starting new bitcoin daemon...');
                await this._start(params);
            }
        }
        await this._init();
        this._ready = true;
        this.emit('ready', this.client);
        return this.client;
    }
    async shutdown() {
        if (!this._closing) {
            this._closing = true;
            if (this._proc !== undefined) {
                const is_dead = this._proc.kill();
                if (!is_dead)
                    this._proc.kill('SIGKILL');
            }
        }
    }
    async run(...methods) {
        if (this._proc === undefined) {
            await this.startup();
        }
        const jobs = methods.map(async (fn) => {
            return Promise.resolve(fn(this.client))
                .then(() => null)
                .catch((err) => err);
        });
        const ret = await Promise.all(jobs);
        await this.shutdown();
        return ret;
    }
}
//# sourceMappingURL=core.js.map